/*
Romu Pseudorandom Number Generators

Copyright 2020 Mark A. Overton

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Website: romu-random.org
Paper:   http://arxiv.org/abs/2002.11331

Copy and paste the generator you want from those below.
To compile, you will need to #include <stdint.h> and use the ROTL definition below.
*/

#include "rtcproto/romu.h"

/*
===== RomuMono32
32-bit arithmetic: Suitable only up to 2^26 output-values. Outputs 16-bit numbers.
Fixed period of (2^32)-47. Must be seeded using the romuMono32_init function.
Capacity = 2^27 bytes. Register pressure = 2. State size = 32 bits.
*/

uint32_t romu_mono32_init(uint32_t seed) {
    return (seed & 0x1fffffffu) + 1156979152u;  /* Accepts 29 seed-bits. */
}

uint16_t romu_mono32_random(uint32_t* state) {
    uint16_t result = *state >> 16;
    *state *= 3611795771u;
    *state = ROTL(*state,12);
    return result;
}
