#include "rtcproto/rtcproto.h"

void rtcp_assert_always(int cond, char* fmt, ...)
{
    va_list argp;
    va_start(argp, fmt);
    if(!cond){
        fprintf(stderr, "rtcp_assert_always: ");
        vfprintf(stderr, fmt, argp);
        exit(1);
    }
    va_end(argp);   
}

unsigned int rtcp_ceil_power_of_two(unsigned int x)
{
    int n = 0;
    while(x > 1){ ++n; x >>= 1; }
    return x << (n + 1);
}

void rtcp_timespec_norm(struct timespec* spec)
{
    spec->tv_sec = spec->tv_sec + (spec->tv_nsec / 1000000000);
    spec->tv_nsec = spec->tv_nsec - (spec->tv_nsec / 1000000000) * 1000000000;
}

int rtcp_timespec_diff_nsec(struct timespec* start, struct timespec* end)
{
    if(start->tv_sec - end->tv_sec){
        return 1000000000 * (end->tv_sec - start->tv_sec - 1) +
               1000000000 + end->tv_nsec - start->tv_nsec;
    } else {
        return end->tv_nsec - start->tv_nsec;
    }
}

int rtcp_timespec_less(struct timespec* start, struct timespec* end)
{
    return (start->tv_sec < end->tv_sec) ?
        1 : (start->tv_sec > end->tv_sec) ?
            0 : (start->tv_nsec < end->tv_nsec);
}

void rtcp_timespec_next(struct timespec* start)
{
    start->tv_sec += (start->tv_nsec == LONG_MAX);
    ++start->tv_nsec;
}

void rtcp_timespec_prev(struct timespec* start)
{
    start->tv_sec -= (start->tv_nsec == 0);
    --start->tv_nsec;
}

struct timespec rtcp_timespec_add(struct timespec* start, struct timespec* end)
{
    struct timespec ret;
    time_t start_sec_from_nsec;
    time_t end_sec_from_nsec;
    start_sec_from_nsec = start->tv_nsec / 1000000000;
    end_sec_from_nsec = end->tv_nsec / 1000000000;
    
    ret.tv_sec = end->tv_sec + start->tv_sec +
                 end_sec_from_nsec + start_sec_from_nsec;
    ret.tv_nsec = (end->tv_nsec - end_sec_from_nsec * 1000000000) +
                  (start->tv_nsec - start_sec_from_nsec * 1000000000);

    return ret;
}

struct timespec rtcp_timespec_diff(struct timespec* start, struct timespec* end)
{
    struct timespec ret;
    time_t sec_diff;
    struct timespec start_norm = *start;
    struct timespec end_norm   = *end;
    rtcp_timespec_norm(&start_norm);
    rtcp_timespec_norm(&end_norm);

    sec_diff = end_norm.tv_sec - start_norm.tv_sec;

    ret.tv_sec = sec_diff - (sec_diff > 0);
    ret.tv_nsec = (sec_diff > 0) * 1000000000 +
                  (end_norm.tv_nsec - start_norm.tv_nsec);

    return ret;
}
