#ifndef TEST_COMMON_H
#define TEST_COMMON_H

#define _POSIX_C_SOURCE 200112L
#include "rtcproto/posixver.h"

#include <unistd.h>

#include "rtcproto/rtcproto.h"
#include "rtcproto/sched.h"
#include "rtcproto/hmap.h"
#include "rtcproto/uvec.h"
#include "rtcproto/heap.h"

typedef struct test_status_t
{
    int   ret;
    char* msg;
} test_status_t;

typedef test_status_t (*fntest)(void);

typedef struct test_entry_t
{
    fntest fun;
    char* name;
    int ret;
} test_entry_t;

#define TEST_FAIL 1
#define TEST_PASS 0

#define TEST_DECLARE(FUN) \
    test_status_t test_ ## FUN(void);
#define TEST_ENTRY(FUN) \
    { test_ ## FUN, #FUN, TEST_FAIL, }

#define TEST_IMPL(FUN) \
    test_status_t test_ ## FUN(void); \
    test_status_t test_ ## FUN()

#define TEST_SUCCESS \
{ \
    test_status_t status; \
    status.ret = TEST_PASS; \
    status.msg = NULL; \
    return status; \
}

#define ASSERT_EXT(expr, message) \
{ \
    if(!(expr)){ \
        test_status_t status; \
        status.ret = TEST_FAIL; \
        status.msg = message; \
        return status; \
    } \
}

#endif /* TEST_COMMON_H */
