#include "../include/common.h"

char fn_proto_1(){ printf("fn_proto_1\n"); return 0; }
char fn_proto_2(){ printf("fn_proto_2\n"); return 0; }
char fn_proto_3(){ printf("fn_proto_3\n"); return 0; }

TEST_IMPL(hmap_add)
{
    int ret;
    int val;

    rtcp_hmap_t hmap;
    ret = rtcp_hmap_malloc(&hmap, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_hmap_init(&hmap);

    val = 0x0BADF00D; ret = rtcp_hmap_add(&hmap, fn_proto_1, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    rtcp_hmap_free(&hmap);

    TEST_SUCCESS
}

TEST_IMPL(hmap_get)
{
    int ret;
    int val;
    void* get;

    rtcp_hmap_t hmap;
    ret = rtcp_hmap_malloc(&hmap, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_hmap_init(&hmap);

    val = 0x0BADF00D; ret = rtcp_hmap_add(&hmap, fn_proto_1, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    get = rtcp_hmap_get(&hmap, fn_proto_1);
    ASSERT_EXT(get && *(int*)get == 0x0BADF00D, "could not retrieve inserted value");

    rtcp_hmap_free(&hmap);

    TEST_SUCCESS
}

TEST_IMPL(hmap_get_multi)
{
    int ret;
    int val;
    void* get;

    rtcp_hmap_t hmap;
    ret = rtcp_hmap_malloc(&hmap, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_hmap_init(&hmap);

    val = 0x0BADF00D; ret = rtcp_hmap_add(&hmap, fn_proto_1, &val);
    ASSERT_EXT(ret == 0, "failed to add item");
    val = 0x0BADC0DE; ret = rtcp_hmap_add(&hmap, fn_proto_2, &val);
    ASSERT_EXT(ret == 0, "failed to add item");
    val = 0x0D15EA5E; ret = rtcp_hmap_add(&hmap, fn_proto_3, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    get = rtcp_hmap_get(&hmap, fn_proto_1);
    ASSERT_EXT(get && *(int*)get == 0x0BADF00D, "could not retrieve inserted value");
    get = rtcp_hmap_get(&hmap, fn_proto_2);
    ASSERT_EXT(get && *(int*)get == 0x0BADC0DE, "could not retrieve inserted value");
    get = rtcp_hmap_get(&hmap, fn_proto_3);
    ASSERT_EXT(get && *(int*)get == 0x0D15EA5E, "could not retrieve inserted value");

    rtcp_hmap_free(&hmap);

    TEST_SUCCESS
}

TEST_IMPL(hmap_del)
{
    int ret;
    int val;
    void *get;

    rtcp_hmap_t hmap;
    ret = rtcp_hmap_malloc(&hmap, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_hmap_init(&hmap);

    val = 0x0BADF00D; ret = rtcp_hmap_add(&hmap, fn_proto_1, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    rtcp_hmap_del(&hmap, fn_proto_1);

    get = rtcp_hmap_get(&hmap, fn_proto_1);
    ASSERT_EXT(!get, "attempt to get deleted value did not fail");

    rtcp_hmap_free(&hmap);

    TEST_SUCCESS
}

TEST_IMPL(hmap_del_multi)
{
    int ret;
    int val;
    void *get;

    rtcp_hmap_t hmap;
    ret = rtcp_hmap_malloc(&hmap, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_hmap_init(&hmap);

    val = 0x0BADF00D; ret = rtcp_hmap_add(&hmap, fn_proto_1, &val);
    ASSERT_EXT(ret == 0, "failed to add item");
    val = 0x0BADC0DE; ret = rtcp_hmap_add(&hmap, fn_proto_2, &val);
    ASSERT_EXT(ret == 0, "failed to add item");
    val = 0x0D15EA5E; ret = rtcp_hmap_add(&hmap, fn_proto_3, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    rtcp_hmap_del(&hmap, fn_proto_1);
    get = rtcp_hmap_get(&hmap, fn_proto_1);
    ASSERT_EXT(!get, "attempt to get deleted value did not fail");

    get = rtcp_hmap_get(&hmap, fn_proto_2);
    ASSERT_EXT(get && *(int*)get == 0x0BADC0DE, "could not retrieve inserted value");
    get = rtcp_hmap_get(&hmap, fn_proto_3);
    ASSERT_EXT(get && *(int*)get == 0x0D15EA5E, "could not retrieve inserted value");

    rtcp_hmap_del(&hmap, fn_proto_2);
    get = rtcp_hmap_get(&hmap, fn_proto_2);
    ASSERT_EXT(!get, "attempt to get deleted value did not fail");

    get = rtcp_hmap_get(&hmap, fn_proto_3);
    ASSERT_EXT(get && *(int*)get == 0x0D15EA5E, "could not retrieve inserted value");

    rtcp_hmap_del(&hmap, fn_proto_3);
    get = rtcp_hmap_get(&hmap, fn_proto_3);
    ASSERT_EXT(!get, "attempt to get deleted value did not fail");

    rtcp_hmap_free(&hmap);

    TEST_SUCCESS
}

TEST_IMPL(hmap_shrink)
{
    int ret;
    int val;
    int i;
    int n, n2;

    rtcp_hmap_t hmap;
    ret = rtcp_hmap_malloc(&hmap, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_hmap_init(&hmap);

    /* Test with n items, where n is a power of two */
    n = n2 = 16;

    for(i = 0; i < n; ++i){
		val = i; ret = rtcp_hmap_add(&hmap, (fn_prthread_t)(unsigned long)i, &val);
		ASSERT_EXT(ret == 0, "failed to add item");
    }
    ASSERT_EXT(hmap.capacity == n * 2, "wrong capacity after initial fill");

    for(i = 0; i < n; ++i){ rtcp_hmap_tick(&hmap); }

    rtcp_hmap_del(&hmap, (fn_prthread_t)(unsigned long)--n2);

    ASSERT_EXT(hmap.capacity == n * 2, "capacity should not have shrunk");

    while(n2 > n / 2){
		rtcp_hmap_del(&hmap, (fn_prthread_t)(unsigned long)--n2);
    }

    ASSERT_EXT(hmap.capacity == n * 2, "capacity should not have shrunk without ticking");
    ASSERT_EXT(hmap.size == n / 2, "exactly half of the elements should have been removed");

    for(i = 0; i < n; ++i){ rtcp_hmap_tick(&hmap); }

    ASSERT_EXT(hmap.capacity == n * 2, "capacity should not have shrunk after ticking but no write op");

    for(i = 0; i < n; ++i){ rtcp_hmap_tick(&hmap); }

    rtcp_hmap_del(&hmap, (fn_prthread_t)(unsigned long)--n2);

    ASSERT_EXT(hmap.capacity == n, "capacity should have shrunk after ticking and write op");

    while(n2 > 0){
		rtcp_hmap_del(&hmap, (fn_prthread_t)(unsigned long)--n2);
		for(i = 0; i < n; ++i){ rtcp_hmap_tick(&hmap); }
    }

    ASSERT_EXT(hmap.capacity == 2, "capacity should have returned to 2 after removing all");

    rtcp_hmap_free(&hmap);

    TEST_SUCCESS
}

