#include "../include/common.h"

typedef struct StringState
{
	int i;
	int n;
	char* str;
} StringState;

PT_THREAD(fn_proto_ascend(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	char str[4];
	StringState* stt = (StringState*)state;
	(void)end;
	
	PT_BEGIN(pt)

	for(stt->i = 0; stt->i < stt->n; ++stt->i){

		sprintf(str, "%d,", stt->i);
		strcat(stt->str, str);
		PT_YIELD(pt);
	}
	
	PT_END(pt)
}

TEST_IMPL(proto_yield)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);
	now = rtcp_timespec_add(&now, &offset);

	state.i = 0;
	state.n = 4;
	state.str = str;

	ret = rtcp_sched_queue(&sched, now, fn_proto_ascend, &state, sizeof(StringState));
	ASSERT_EXT(ret == 0, "failed to allocate");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued calls");
	ASSERT_EXT(strcmp(str, "0,") == 0, "wrong result string");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,") == 0, "wrong result string");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,2,") == 0, "wrong result string");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,2,3,") == 0, "wrong result string");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

TEST_IMPL(proto_yield_flush)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);
	now = rtcp_timespec_add(&now, &offset);

	state.i = 0;
	state.n = 8;
	state.str = str;

	ret = rtcp_sched_queue(&sched, now, fn_proto_ascend, &state, sizeof(StringState));
	ASSERT_EXT(ret == 0, "failed to allocate");

	rtcp_sched_flush(&sched);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,2,3,4,5,6,7,") == 0, "wrong result string");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

typedef struct StringStateChild
{
	int i;
	int n;
	char* str;
	rtcp_pt_t pt;
} StringStateChild;

PT_THREAD(fn_proto_ascend_early_exit(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	char str[4];
	StringStateChild* stt = (StringStateChild*)state;
	(void)end;
	
	PT_BEGIN(pt)
	
	stt->i = 0;
	while(stt->i < stt->n){
		
		sprintf(str, "%d,", stt->i);
		strcat(stt->str, str);
		++stt->i;
		if(stt->i == 2){ PT_EXIT(pt); }
		else { PT_YIELD(pt); }
	}
	
	PT_END(pt)
}

TEST_IMPL(proto_exit)
{
	int ret;
	char str[8*4] = {0};
	StringStateChild state;
	struct timespec now;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);
	now = rtcp_timespec_add(&now, &offset);

	state.i = 0;
	state.n = 8;
	state.str = str;

	ret = rtcp_sched_queue(&sched, now, fn_proto_ascend_early_exit, &state, sizeof(StringStateChild));
	ASSERT_EXT(ret == 0, "failed to allocate");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued calls");
	ASSERT_EXT(strcmp(str, "0,") == 0, "wrong result string");
	
	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,") == 0, "wrong result string");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

PT_THREAD(fn_proto_ascend_spawn(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	char str[4];
	StringStateChild* stt = (StringStateChild*)state;
	(void)end;
	
	PT_BEGIN(pt)

	PT_SPAWN(pt, &stt->pt, fn_proto_ascend_early_exit(&stt->pt, state, end));
	PT_YIELD(pt);
	while(stt->i < stt->n){

		sprintf(str, "%d,", stt->i);
		strcat(stt->str, str);
		++stt->i;
	}
	
	PT_END(pt)
}

TEST_IMPL(proto_spawn)
{
	int ret;
	char str[8*4] = {0};
	StringStateChild state;
	struct timespec now;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);
	now = rtcp_timespec_add(&now, &offset);

	state.i = 0;
	state.n = 8;
	state.str = str;

	ret = rtcp_sched_queue(&sched, now, fn_proto_ascend_spawn, &state, sizeof(StringStateChild));
	ASSERT_EXT(ret == 0, "failed to allocate");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued calls");
	ASSERT_EXT(strcmp(str, "0,") == 0, "wrong result string");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,") == 0, "wrong result string");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,2,3,4,5,6,7,") == 0, "wrong result string");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

PT_THREAD(fn_proto_ascend_coop(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	char str[4];
	StringState* stt = (StringState*)state;
	(void)end;
	
	PT_BEGIN(pt)

	stt->i = 0;

	sprintf(str, "%d,", stt->i++);
	strcat(stt->str, str);

	PT_YIELD_COOP(pt, end);

	while(stt->i < stt->n){
		sprintf(str, "%d,", stt->i);
		strcat(stt->str, str);
		++stt->i;
	}
	
	PT_END(pt)
}

TEST_IMPL(proto_yield_noncoop)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);

	state.i = 0;
	state.n = 4;
	state.str = str;

	ret = rtcp_sched_queue(&sched, now, fn_proto_ascend_coop, &state, sizeof(StringState));
	ASSERT_EXT(ret == 0, "failed to allocate");

	rtcp_sched_call(&sched, now, 1);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,2,3,") == 0, "wrong result string");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

TEST_IMPL(proto_yield_coop)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);

	state.i = 0;
	state.n = 4;
	state.str = str;

	ret = rtcp_sched_queue(&sched, now, fn_proto_ascend_coop, &state, sizeof(StringState));
	ASSERT_EXT(ret == 0, "failed to allocate");

	rtcp_sched_call(&sched, now, 0);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued calls");
	ASSERT_EXT(strcmp(str, "0,") == 0, "wrong result string");

	now = rtcp_timespec_add(&now, &offset);

	rtcp_sched_call(&sched, now, 0);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,2,3,") == 0, "wrong result string");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

PT_THREAD(fn_proto_loop_coop(rtcp_pt_t* pt, void* state, struct timespec* end))
{
	char str[4];
	StringState* stt = (StringState*)state;
	int i;
	(void)end;
	
	PT_BEGIN(pt)
	
	PT_LOOP(stt->i = 0, stt->i < stt->n, i, 2){

		sprintf(str, "%d,", stt->i);
		strcat(stt->str, str);

		PT_LOOP_INCR(pt, end, ++stt->i, i)
	}
	
	PT_END(pt)
}

TEST_IMPL(proto_loop_coop)
{
	int ret;
	char str[8*4] = {0};
	StringState state;
	struct timespec now;
	struct timespec offset = { 0, 10000000 };

	rtcp_sched_t sched;
	ret = rtcp_sched_malloc(&sched);
	ASSERT_EXT(ret == 0, "failed to allocate");

	clock_gettime(CLOCK_REALTIME, &now);

	state.i = 0;
	state.n = 4;
	state.str = str;

	ret = rtcp_sched_queue(&sched, now, fn_proto_loop_coop, &state, sizeof(StringState));
	ASSERT_EXT(ret == 0, "failed to allocate");

	rtcp_sched_call(&sched, now, 0);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 1, "expected 1 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,") == 0, "wrong result string");

	now = rtcp_timespec_add(&now, &offset);

	rtcp_sched_call(&sched, now, 0);
	ASSERT_EXT(rtcp_sched_calls_queued(&sched) == 0, "expected 0 queued calls");
	ASSERT_EXT(strcmp(str, "0,1,2,3,") == 0, "wrong result string");

	rtcp_sched_free(&sched);

	TEST_SUCCESS
}

