#include "../include/common.h"

void uvec_callback_1(void* val, int idx, void* user)
{
    (void)val;
    (void)idx;

    /* Increment flag */
    ++(*(int*)user);
}

void uvec_callback_2(void* val, int idx, void* user)
{
    int* expect;

    (void)val;
    
    /* Get expected index */
    expect = (int*)user;
    
    /* Check that the index moved into was the expected one */
    *expect = (*expect == idx);
}

TEST_IMPL(uvec_add)
{
    int ret;
    int val;

    rtcp_uvec_t uvec;
    ret = rtcp_uvec_malloc(&uvec, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_uvec_init(&uvec);

    val = 0x0BADF00D; ret = rtcp_uvec_add(&uvec, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    rtcp_uvec_free(&uvec);

    TEST_SUCCESS
}

TEST_IMPL(uvec_add_callback)
{
    int ret;
    int val;
    int flag = 0;

    rtcp_uvec_t uvec;
    ret = rtcp_uvec_malloc(&uvec, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_uvec_init(&uvec);
    rtcp_uvec_cb(&uvec, uvec_callback_1, &flag);

    val = 0x0BADF00D; ret = rtcp_uvec_add(&uvec, &val);
    ASSERT_EXT(ret == 0, "failed to add item");
    ASSERT_EXT(flag == 1, "callback counter expected 1");

    rtcp_uvec_free(&uvec);

    TEST_SUCCESS
}

TEST_IMPL(uvec_del)
{
    int ret;
    int val;

    rtcp_uvec_t uvec;
    ret = rtcp_uvec_malloc(&uvec, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_uvec_init(&uvec);

    val = 0x0BADF00D; ret = rtcp_uvec_add(&uvec, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    ret = rtcp_uvec_del(&uvec, 0);
    ASSERT_EXT(ret == 0, "failed to delete item");

    rtcp_uvec_free(&uvec);

    TEST_SUCCESS
}

TEST_IMPL(uvec_del_callback)
{
    int ret;
    int val;
    int expect;
    int flag;

    rtcp_uvec_t uvec;
    ret = rtcp_uvec_malloc(&uvec, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_uvec_init(&uvec);
    rtcp_uvec_cb(&uvec, uvec_callback_2, &expect);

    val = 0x0BADF00D; ret = rtcp_uvec_add(&uvec, &val);
    ASSERT_EXT(ret == 0, "failed to add item");
    val = 0x0BADC0DE; ret = rtcp_uvec_add(&uvec, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    expect = 0;
    ret = rtcp_uvec_del(&uvec, 0);
    ASSERT_EXT(ret == 0, "failed to delete item");
    ASSERT_EXT(expect != 0, "callback tester returned false (zero)");

    rtcp_uvec_cb(&uvec, uvec_callback_1, &flag);
    flag = 0;
    ret = rtcp_uvec_del(&uvec, 0);
    ASSERT_EXT(ret == 0, "failed to delete item");
    ASSERT_EXT(flag == 0, "callback should not have been called after last delete");

    rtcp_uvec_free(&uvec);

    TEST_SUCCESS
}

TEST_IMPL(uvec_del_callback_last)
{
    int ret;
    int val;
    int flag;

    rtcp_uvec_t uvec;
    ret = rtcp_uvec_malloc(&uvec, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_uvec_init(&uvec);
    rtcp_uvec_cb(&uvec, uvec_callback_1, &flag);

    val = 0x0BADF00D; ret = rtcp_uvec_add(&uvec, &val);
    ASSERT_EXT(ret == 0, "failed to add item");
    val = 0x0BADC0DE; ret = rtcp_uvec_add(&uvec, &val);
    ASSERT_EXT(ret == 0, "failed to add item");

    flag = 0;
    ret = rtcp_uvec_del(&uvec, 1);
    ASSERT_EXT(ret == 0, "failed to delete item");
    ASSERT_EXT(flag == 0, "callback should not have been called when the last element is deleted");

    flag = 0;
    ret = rtcp_uvec_del(&uvec, 0);
    ASSERT_EXT(ret == 0, "failed to delete item");
    ASSERT_EXT(flag == 0, "callback should not have been called when the last element is deleted");

    rtcp_uvec_free(&uvec);

    TEST_SUCCESS
}

TEST_IMPL(uvec_shrink)
{
    int ret;
    int val;
    int i;
    int n, n2;
    
    rtcp_uvec_t uvec;
    ret = rtcp_uvec_malloc(&uvec, 1, sizeof(int));
    ASSERT_EXT(ret == 0, "failed to allocate");
    rtcp_uvec_init(&uvec);

    /* Test with n items, where n is a power of two */
    n = n2 = 16;

    for(i = 0; i < n; ++i){
		val = i; ret = rtcp_uvec_add(&uvec, &val);
		ASSERT_EXT(ret == 0, "failed to add item");
    }
    ASSERT_EXT(uvec.capacity == n, "wrong capacity after initial fill");

    for(i = 0; i < n; ++i){ rtcp_uvec_tick(&uvec); }

    ret = rtcp_uvec_del(&uvec, --n2);
    ASSERT_EXT(ret == 0, "failed to delete item");

    ASSERT_EXT(uvec.capacity == n, "capacity should not have shrunk below used size");

    while(n2 > n / 2){
		ret = rtcp_uvec_del(&uvec, --n2);
		ASSERT_EXT(ret == 0, "failed to delete item");
    }

    ASSERT_EXT(uvec.capacity == n, "capacity should not have shrunk without ticking");
    ASSERT_EXT(uvec.size == n / 2, "exactly half of the elements should have been removed");

    for(i = 0; i < n; ++i){ rtcp_uvec_tick(&uvec); }

    ASSERT_EXT(uvec.capacity == n, "capacity should not have shrunk after ticking but no write op");

    for(i = 0; i < n; ++i){ rtcp_uvec_tick(&uvec); }

    ret = rtcp_uvec_del(&uvec, --n2);
    ASSERT_EXT(ret == 0, "failed to delete item");

    ASSERT_EXT(uvec.capacity == n / 2, "capacity should have shrunk after ticking and write op");

    while(n2 > 0){
		ret = rtcp_uvec_del(&uvec, --n2);
		ASSERT_EXT(ret == 0, "failed to delete item");
		for(i = 0; i < n; ++i){ rtcp_uvec_tick(&uvec); }
    }

    ASSERT_EXT(uvec.capacity == 1, "capacity should have returned to 1 after removing all");

    rtcp_uvec_free(&uvec);

    TEST_SUCCESS
}










